#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import enum
import json
import logging
import os
import sys
from typing import Callable, Union
from abc import ABC, abstractmethod
from http import HTTPStatus
from urllib.parse import urlparse

import requests
from google.auth.transport.requests import Request
from google.cloud import storage
from google.oauth2 import service_account
from libs.exceptions import RefreshSATokenError, SAFilePathError
from tenacity import retry, stop_after_attempt

# Set up base logger
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(
    logging.Formatter("%(asctime)s [%(name)-14.14s] [%(levelname)-7.7s]  %(message)s"))
logger = logging.getLogger("Dataload")
logger.setLevel(logging.INFO)
logger.addHandler(handler)

RETRIES = 3


class TokenRefresher(ABC):

    @abstractmethod
    def refresh_token(self) -> str:
        """
        Implement logics of refreshing token here.
        """
        pass

    @property
    @abstractmethod
    def access_token(self) -> str:
        pass

    @property
    @abstractmethod
    def authorization_header(self) -> dict:
        """
        Must return  authorization header for updating headers dict.
        E.g. return {"Authorization": "Bearer <access_token>"}
        """
        pass


class AirflowTokenRefresher(TokenRefresher):
    DEFAULT_ACCESS_SCOPES = ['openid', 'email', 'profile']

    def __init__(self, access_scopes: list=None):
        from airflow.models import Variable
        self.airflow_variables = Variable
        self._access_token = None
        self._access_scopes = access_scopes

    @property
    def access_scopes(self) -> list:
        """
        Return access scopes.
        Use DEFAULT_ACCESS_SCOPES if user-defined ones weren't provided.
        """
        if not self._access_scopes:
            self._access_scopes = self.DEFAULT_ACCESS_SCOPES
        return self._access_scopes

    @staticmethod
    @retry(stop=stop_after_attempt(RETRIES))
    def get_sa_file_content_from_google_storage(bucket_name: str, source_blob_name: str) -> str:
        """
        Get sa_file content from Google Storage.
        """
        storage_client = storage.Client()
        bucket = storage_client.bucket(bucket_name)
        blob = bucket.blob(source_blob_name)
        logger.info("Got SA_file.")
        return blob.download_as_string()

    def get_sa_file_info(self) -> dict:
        """
        Get file path from SA_FILE_PATH environmental variable.
        This path can be GCS object URI or local file path.
        Return content of sa path as dict.
        """
        sa_file_path = os.environ.get("SA_FILE_PATH", None)
        parsed_path = urlparse(sa_file_path)
        if parsed_path.scheme == "gs":
            bucket_name = parsed_path.netloc
            source_blob_name = parsed_path.path[1:]  # delete the first slash
            sa_file_content = self.get_sa_file_content_from_google_storage(bucket_name,
                                                                           source_blob_name)
            sa_file_info = json.loads(sa_file_content)
        elif not parsed_path.scheme and os.path.isfile(parsed_path.path):
            with open(parsed_path.path) as f:
                sa_file_info = json.load(f)
        else:
            raise SAFilePathError
        return sa_file_info

    @retry(stop=stop_after_attempt(RETRIES))
    def get_access_token_using_sa_file(self) -> str:
        """
        Get new access token using SA info.
        """
        sa_file_content = self.get_sa_file_info()
        try:
            credentials = service_account.Credentials.from_service_account_info(
                sa_file_content, scopes=self.access_scopes)
        except ValueError as e:
            logger.error("SA file has bad format.")
            raise e

        logger.info("Refresh token.")
        credentials.refresh(Request())
        token = credentials.token

        if credentials.token is None:
            logger.error("Can't refresh token using SA-file. Token is empty.")
            raise RefreshSATokenError

        logger.info("Token refreshed.")
        return token

    @retry(stop=stop_after_attempt(RETRIES))
    def refresh_token(self) -> str:
        """
        Refresh token. Store its value in Airflow Variable.
        """
        token = self.get_access_token_using_sa_file()
        self.airflow_variables.set("access_token", token)
        self._access_token = token
        return self._access_token

    @property
    def access_token(self) -> str:
        if not self._access_token:
            try:
                self._access_token = self.airflow_variables.get("access_token")
            except KeyError:
                self.refresh_token()
        return self._access_token

    @property
    def authorization_header(self) -> dict:
        return {"Authorization": f"Bearer {self.access_token}"}


def make_callable_request(obj: Union[object, None], request_function: Callable, headers: dict,
                          *args, **kwargs) -> Callable:
    """
    Create send_request_with_auth function.
    """
    if obj:  # if wrapped function is an object's method
        callable_request = lambda: request_function(obj, headers, *args, **kwargs)
    else:
        callable_request = lambda: request_function(headers, *args, **kwargs)
    return callable_request


def _validate_headers_type(headers: dict):
    if not isinstance(headers, dict):
        logger.error(f"Got headers {headers}")
        raise TypeError


def _validate_response_type(response: requests.Response, request_function: Callable):
    if not isinstance(response, requests.Response):
        logger.error(f"Function or method {request_function}"
                     f" must return values of type 'requests.Response'. "
                     f"Got {type(response)} instead")
        raise TypeError


def _validate_token_refresher_type(token_refresher: TokenRefresher):
    if not isinstance(token_refresher, TokenRefresher):
        raise TypeError(
            f"Token refresher must be of type {TokenRefresher}. Got {type(token_refresher)}"
        )


def send_request_with_auth_header(token_refresher: TokenRefresher, *args,
                                   **kwargs) -> requests.Response:
    """
    Send request with authorization token. If response status is in HTTPStatus.UNAUTHORIZED or
    HTTPStatus.FORBIDDEN, then refresh token and send request once again.
    """
    obj = kwargs.pop("obj") if kwargs.get("obj") else None
    request_function = kwargs.pop("request_function")
    headers = kwargs.pop("headers")
    _validate_headers_type(headers)
    headers.update(token_refresher.authorization_header)

    send_request_with_auth = make_callable_request(obj, request_function, headers,
                                                   *args, **kwargs)
    response = send_request_with_auth()
    _validate_response_type(response, request_function)

    if not response.ok:
        if response.status_code in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
            token_refresher.refresh_token()
            headers.update(token_refresher.authorization_header)
            send_request_with_auth = make_callable_request(obj,
                                                           request_function,
                                                           headers,
                                                           *args, **kwargs)
            response = send_request_with_auth()
        response.raise_for_status()
    return response


def refresh_token(token_refresher: TokenRefresher) -> Callable:
    """
    Wrap a request function and check response. If response's error status code
    is about Authorization, refresh token and invoke this function once again.
    Expects function:
    If response is not ok and not about Authorization, then raises HTTPError
    request_func(header: dict, *args, **kwargs) -> requests.Response
    Or method:
    request_method(self, header: dict, *args, **kwargs) -> requests.Response
    """
    _validate_token_refresher_type(token_refresher)

    def refresh_token_wrapper(request_function: Callable) -> Callable:
        is_method = len(request_function.__qualname__.split(".")) > 1
        if is_method:
            def _wrapper(obj: object, headers: dict, *args, **kwargs) -> requests.Response:
                return send_request_with_auth_header(token_refresher,
                                                      request_function=request_function,
                                                      obj=obj,
                                                      headers=headers,
                                                      *args,
                                                      **kwargs)
        else:
            def _wrapper(headers: dict, *args, **kwargs) -> requests.Response:
                return send_request_with_auth_header(token_refresher,
                                                      request_function=request_function,
                                                      headers=headers,
                                                      *args, **kwargs)
        return _wrapper

    return refresh_token_wrapper
