#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import dataclasses
import json
import logging
import sys
import uuid
from typing import List

import jsonschema
import requests
import tenacity
from airflow.models import BaseOperator, Variable
from libs.exceptions import EmptyManifestError, NotOSDUShemaFormatError
from libs.refresh_token import AirflowTokenRefresher, refresh_token

# Set up base logger
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(
    logging.Formatter("%(asctime)s [%(name)-14.14s] [%(levelname)-7.7s]  %(message)s"))
logger = logging.getLogger("Dataload")
logger.setLevel(logging.INFO)
logger.addHandler(handler)

RETRIES = 3
TIMEOUT = 1


@dataclasses.dataclass
class Context(object):
    """
    Store data-partition-id and AppKey passed via Payload field of dagrun.conf.
    Remove Payload from dagrun.conf.
    """
    data_partition_id: str
    app_key: str

    @classmethod
    def populate(cls, ctx: dict) -> 'Context':
        ctx_payload = ctx.pop('Payload')
        app_key = ctx_payload['AppKey']
        data_partition_id = ctx_payload['data-partition-id']
        ctx_obj = cls(app_key=app_key, data_partition_id=data_partition_id)
        return ctx_obj


class OSDURefResolver(jsonschema.RefResolver):

    def __init__(self, schema_service: str, *args, **kwargs):
        super(OSDURefResolver, self).__init__(*args, **kwargs)
        self.schema_service = schema_service

    def resolve_fragment(self, document: dict, fragment: str) -> dict:
        """
        Extend base resolve_fragment method. If a nested schema has 'definitions' field and there
        is a schema under this 'definitions', jsonschema attempts to use the id field of this
        double-nested schema as URI to get this schema later. So it has sense to replace this id
        with a correct one.
        """
        document = super().resolve_fragment(document, fragment)
        fragment_parts = fragment.split("/")  # /definitions/<OsduID> -> [..., <OsduID>]
        if len(fragment_parts) > 1:
            osdu_id = fragment_parts[-1]
            url = f"{self.schema_service}/{osdu_id}"
            document["$id"] = url
        return document


class SchemaValidator(object):
    """Class to validate schema of Manifests."""

    def __init__(self, schema_service: str, dagrun_conf: dict, context: Context):
        self.schema_service = schema_service
        self.data_object = copy.deepcopy(dagrun_conf)
        self.context = context
        self.resolver_handlers = {"osdu": self.get_schema_request,
                                  "https": self.get_schema_request,
                                  self.context.data_partition_id: self.get_schema_request}
        self.create_request_headers()

    def create_request_headers(self):
        self.request_headers = {
            'Content-type': 'application/json',
            'data-partition-id': self.context.data_partition_id,
            'AppKey': self.context.app_key,
        }

    @refresh_token(AirflowTokenRefresher())
    def _get_schema_request(self, headers: dict, uri: str) -> requests.Response:
        """
        Request to Schema service to retrieve schema.
        """
        response = requests.get(uri, headers=headers, timeout=60)
        return response

    def get_schema_request(self, uri: str) -> dict:
        """
        Get Schema from Schema service. Change $id field to url of getting schema.
        """
        if uri.startswith("osdu") or uri.startswith(self.context.data_partition_id):
            uri = f"{self.schema_service}/{uri}"
        response = self._get_schema_request(self.request_headers, uri).json()
        response["$id"] = uri
        return response

    def get_schema(self, kind: str) -> dict:
        manifest_schema_uri = f"{self.schema_service}/{kind}"
        try:
            response = self.get_schema_request(manifest_schema_uri)
        except Exception as e:
            logger.error(f"Error on getting schema of kind '{kind}'")
            raise e
        return response

    def _validate_schema(self, manifest: dict, schema: dict = None):
        """
        Validate schema. If argument schema is not defined, then use schema service to retrieve
        corresponding schema.
        """
        if not schema:
            schema = self.get_schema(manifest["kind"])
        logger.info(f"Validating kind {manifest['kind']}")
        resolver = OSDURefResolver(schema_service=self.schema_service,
                                   base_uri=schema.get("$id", ""), referrer=schema,
                                   handlers=self.resolver_handlers, cache_remote=True)
        validator = jsonschema.Draft7Validator(schema=schema, resolver=resolver)
        validator.validate(manifest)

    def validate_work_product(self, work_product: dict):
        """
        Validate WP manifest. Raise error if manifest is not valid.
        """
        for key, value in work_product.items():
            if key != "WorkProduct":
                for component in value:
                    self._validate_schema(component)
            else:
                self._validate_schema(value)

    def validate_manifest(self):
        """
        Validate manifest. Raise error if manifest is not valid.
        """
        for manifest in self.data_object["manifest"]:
            if isinstance(manifest, dict) and manifest.get("kind"):
                self._validate_schema(manifest)
            elif manifest.get("WorkProductComponents"):
                self.validate_work_product(manifest)
            else:
                raise NotOSDUShemaFormatError(f"Not valid schema {manifest}")


class ManifestProcessor(object):
    """Class to process WP, Master and Reference data"""
    RECORD_TEMPLATE = {
        "legal": {},
        "acl": {},
        "kind": "",
        "id": "",
        "data": {
        }
    }

    def __init__(self, storage_url, dagrun_conf, context):
        self.storage_url = storage_url
        self.data_object = copy.deepcopy(dagrun_conf)
        self.context = context

    @staticmethod
    def _get_kind_name(kind: str) -> str:
        """
        osdu:osdu:Well:1.0.0 -> Well
        """
        kind_name = kind.split(":")[2]
        return kind_name

    def generate_id(self, manifest_fragment: dict) -> str:
        """
        Generate id to use it in Storage.
        """
        group_type = manifest_fragment.get("groupType", "doc")
        kind = manifest_fragment.get("kind")
        kind_name = self._get_kind_name(kind)
        _id = f"{self.context.data_partition_id}:{group_type}_{kind_name}:{str(uuid.uuid4())}"
        return _id

    @property
    def request_headers(self) -> dict:
        headers = {
            'Content-type': 'application/json',
            'data-partition-id': self.context.data_partition_id,
            'AppKey': self.context.app_key
        }
        return headers

    def populate_manifest_storage_record(self, manifest: dict) -> dict:
        """
        Create a record from Master-manifest to store it in Storage service
        """
        record = copy.deepcopy(self.RECORD_TEMPLATE)
        record["id"] = self.generate_id(manifest)
        record["kind"] = manifest.pop("kind")
        record["legal"] = manifest.pop("legal")
        record["acl"] = manifest.pop("acl")
        record["data"] = manifest
        return record

    @tenacity.retry(tenacity.wait_fixed(TIMEOUT),
                    tenacity.stop_after_attempt(RETRIES))
    @refresh_token(AirflowTokenRefresher())
    def save_record(self, headers: dict, request_data: List[dict]) -> requests.Response:
        """
        Send request to record storage API.
        """
        response = requests.put(self.storage_url, json.dumps(request_data), headers=headers)
        if response.ok:
            logger.info(",".join(map(str, response.json()["recordIds"])))
        else:
            reason = response.text[:250]
            logger.error(f"Request error.")
            logger.error(f"Response status: {response.status_code}. "
                         f"Response content: {reason}.")
        return response

    def process_work_product(self, manifest: dict) -> List[dict]:
        """
        Process WP.
        """
        wp = manifest["WorkProduct"]
        records = [self.populate_manifest_storage_record(wp)]
        return records

    def process_work_product_components(self, manifest: dict) -> List[dict]:
        """
        Process list of WPS.
        """
        records = []
        for wpc in manifest["WorkProductComponents"]:
            record = self.populate_manifest_storage_record(wpc)
            records.append(record)
        return records

    def process_work_product_files(self, manifest: dict) -> List[dict]:
        """
        Process list of files.
        """
        records = []
        for file in manifest["Files"]:
            record = self.populate_manifest_storage_record(file)
            records.append(record)
        return records

    def process_work_product_manifest(self, manifest: dict) -> List[dict]:
        file_records = self.process_work_product_files(manifest)
        wpc_records = self.process_work_product_components(manifest)
        wp_records = self.process_work_product(manifest)
        records = file_records + wpc_records + wp_records
        return records

    def create_manifest_records(self) -> List[dict]:
        manifest_records = []
        manifests = self.data_object["manifest"]
        for manifest in manifests:
            if "WorkProduct" in manifest:
                wp_records = self.process_work_product_manifest(manifest)
                manifest_records.extend(wp_records)
            else:
                record = self.populate_manifest_storage_record(manifest)
                manifest_records.append(record)
        return manifest_records

    def process_manifest(self) -> List[str]:
        if "manifest" in self.data_object:
            manifest_records = self.create_manifest_records()
        else:
            raise EmptyManifestError
        response = self.save_record(self.request_headers, request_data=manifest_records)
        record_ids = response.json()["recordIds"]
        return record_ids


class ProcessManifestOperatorR3(BaseOperator):
    ui_color = '#dad5ff'
    ui_fgcolor = '#000000'

    def pre_execute(self, context: dict):
        self.schema_service_url = Variable.get('schema_service_url')
        self.storage_url = Variable.get('storage_url')

    def execute(self, context: dict):
        payload_context = Context.populate(context["dag_run"].conf)
        validator = SchemaValidator(
            self.schema_service_url,
            context["dag_run"].conf,
            payload_context
        )
        manifest_processor = ManifestProcessor(
            self.storage_url,
            context["dag_run"].conf,
            payload_context
        )
        validator.validate_manifest()
        record_ids = manifest_processor.process_manifest()
        context["ti"].xcom_push(key="record_ids", value=record_ids)
