#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import json
import logging
import sys
from typing import Tuple

import tenacity
from airflow.models import BaseOperator, Variable
from airflow.utils.decorators import apply_defaults
from hooks import search_http_hook, workflow_hook
from libs.exceptions import RecordsNotSearchableError
from libs.refresh_token import AirflowTokenRefresher, refresh_token

# Set up base logger
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(
    logging.Formatter("%(asctime)s [%(name)-14.14s] [%(levelname)-7.7s]  %(message)s"))
logger = logging.getLogger("Dataload")
logger.setLevel(logging.INFO)
logger.addHandler(handler)


class SearchRecordIdOperator(BaseOperator):
    """
    Operator to search files in SearchService by record ids.
    Expects "record_ids" field in xcom.
    """
    ui_color = '#10ECAA'
    ui_fgcolor = '#000000'

    FINISHED_STATUS = "finished"
    RUNNING_STATUS = "running"
    FAILED_STATUS = "failed"

    @apply_defaults
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.workflow_hook = workflow_hook
        self.search_hook = search_http_hook
        # the will be set at the beginning of execute method
        self.request_body = None
        self.expected_total_count = None

    def get_headers(self, **kwargs) -> dict:
        data_conf = kwargs['dag_run'].conf
        # for /submitWithManifest authorization and partition-id are inside Payload field
        if "Payload" in data_conf:
            partition_id = data_conf["Payload"]["data-partition-id"]
        else:
            partition_id = data_conf["data-partition-id"]
        headers = {
            'Content-type': 'application/json',
            'data-partition-id': partition_id,
            'Authorization': "",
        }
        return headers

    @staticmethod
    def _create_search_query(record_ids) -> Tuple[str, int]:
        expected_total_count = len(record_ids)
        record_ids = " OR ".join(f"\"{id_}\"" for id_ in record_ids)
        logger.info(f"Search query {record_ids}")
        query = f"id:({record_ids})"
        return query, expected_total_count

    def _create_request_body(self, **kwargs):
        record_ids = kwargs["ti"].xcom_pull(key="record_ids", )
        if record_ids:
            query, expected_total_count = self._create_search_query(record_ids)
        else:
            logger.error("There are no record ids")
            sys.exit(2)
        request_body = {
            "kind": "*:*:*:*",
            "query": query
        }
        return request_body, expected_total_count

    def _is_record_searchable(self, resp) -> bool:
        """
        Check if search service returns expected totalCount of records.
        """
        logger.info(resp.text)
        data = resp.json()
        return data.get("totalCount") == self.expected_total_count

    @refresh_token(AirflowTokenRefresher())
    def search_files(self, headers, **kwargs):
        if self.request_body:
            response = self.search_hook.run(
                endpoint=Variable.get("search_query_ep"),
                headers=headers,
                data=json.dumps(self.request_body),
                extra_options={"check_response": False}
            )
            if not self._is_record_searchable(response):
                logger.error("Expected amount (%s) of records not found." %
                             self.expected_total_count
                             )
                raise RecordsNotSearchableError
            return response
        else:
            logger.error("There is an error in header or in request body")
            sys.exit(2)

    def execute(self, context):
        """Execute update workflow status.
           If status assumed to be FINISHED then we check whether proceed files are searchable or not.
           If they are then update status FINISHED else FAILED
        """
        self.request_body, self.expected_total_count = self._create_request_body(**context)
        headers = self.get_headers(**context)
        self.search_files(headers, **context)
