#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import enum
import json
import logging
import sys
from functools import partial

import tenacity
from airflow.models import BaseOperator, Variable
from airflow.utils.decorators import apply_defaults
from hooks.http_hooks import search_http_hook, workflow_hook
from libs.exceptions import PipelineFailedError
from libs.refresh_token import AirflowTokenRefresher, refresh_token

# Set up base logger
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(
    logging.Formatter("%(asctime)s [%(name)-14.14s] [%(levelname)-7.7s]  %(message)s"))
logger = logging.getLogger("Dataload")
logger.setLevel(logging.INFO)
logger.addHandler(handler)


class UpdateStatusOperator(BaseOperator):
    ui_color = '#10ECAA'
    ui_fgcolor = '#000000'

    FINISHED_STATUS = "finished"
    RUNNING_STATUS = "running"
    FAILED_STATUS = "failed"

    class prev_ti_state(enum.Enum):
        NONE = enum.auto()
        SUCCESS = enum.auto()
        FAILED = enum.auto()

    @apply_defaults
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.workflow_hook = workflow_hook
        self.search_hook = search_http_hook

    @staticmethod
    def _file_searched(resp, expected_total_count):
        """Check if search service returns totalCount.
        The method is used  as a callback
        """
        data = resp.json()
        return data.get("totalCount") == expected_total_count

    def get_headers(self, **kwargs):
        data_conf = kwargs['dag_run'].conf
        # for /submitWithManifest authorization and partition-id are inside Payload field
        if "Payload" in data_conf:
            partition_id = data_conf["Payload"]["data-partition-id"]
        else:
            partition_id = data_conf["data-partition-id"]
        headers = {
            'Content-type': 'application/json',
            'data-partition-id': partition_id,
        }
        return headers

    def search_files(self, **kwargs):
        def create_query(record_ids):
            expected_total_count = len(record_ids)
            record_ids = " OR ".join(f"\"{id_}\"" for id_ in record_ids)
            logger.info(f"Search query {record_ids}")
            query = f"id:({record_ids})"
            return query, expected_total_count

        record_ids = kwargs["ti"].xcom_pull(key="record_ids", task_ids='create_records')
        if record_ids:
            query, expected_total_count = create_query(record_ids)
        else:
            logger.error("There are no record ids")
            sys.exit(2)
        headers = self.get_headers(**kwargs)
        request_body = {
            "kind": "*:*:*:*",
            "query": query
        }
        retry_opts = {
            "wait": tenacity.wait_exponential(multiplier=5),
            "stop": tenacity.stop_after_attempt(5),
            "retry": tenacity.retry_if_not_result(
                partial(self._file_searched, expected_total_count=expected_total_count)
            )
        }
        self.search_hook.run_with_advanced_retry(
            endpoint=Variable.get("search_query_ep"),
            headers=headers,
            data=json.dumps(request_body),
            _retry_args=retry_opts
        )

    def previous_ti_statuses(self, context):
        dagrun = context['ti'].get_dagrun()
        failed_ti, success_ti = dagrun.get_task_instances(
            state='failed'), dagrun.get_task_instances(state='success')
        if not failed_ti and not success_ti:  # There is no prev task so it can't have been failed
            logger.info("There are no tasks before this one. So it has status RUNNING")
            return self.prev_ti_state.NONE
        if failed_ti:
            logger.info("There are failed tasks before this one. So it has status FAILED")
            return self.prev_ti_state.FAILED
        logger.info("There are successed tasks before this one. So it has status SUCCESSED")
        return self.prev_ti_state.SUCCESS

    def pre_execute(self, context):
        prev_tis = self.previous_ti_statuses(context)
        if prev_tis is self.prev_ti_state.NONE:
            self.status = self.RUNNING_STATUS
        elif prev_tis is self.prev_ti_state.FAILED:
            self.status = self.FAILED_STATUS
        elif prev_tis is self.prev_ti_state.SUCCESS:
            self.status = self.FINISHED_STATUS

    def execute(self, context):
        """Execute update workflow status.
           If status assumed to be FINISHED then we check whether proceed files are searchable or not.
           If they are then update status FINISHED else FAILED
        """
        headers = self.get_headers(**context)
        self.update_status_request(headers, self.status, **context)
        if self.status == self.FAILED_STATUS:
            raise PipelineFailedError("Dag failed")

    @refresh_token(AirflowTokenRefresher())
    def update_status_request(self, headers, status, **kwargs):
        data_conf = kwargs['dag_run'].conf
        logger.info(f"Got dataconf {data_conf}")
        workflow_id = data_conf["WorkflowID"]
        request_body = {
            "WorkflowID": workflow_id,
            "Status": status
        }
        logger.info(f" Sending request '{status}'")
        response = self.workflow_hook.run(
            endpoint=Variable.get("update_status_ep"),
            data=json.dumps(request_body),
            headers=headers,
            extra_options={"check_response": False}
        )
        return response
